@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2014-2016 Brazil
+  Copyright(C) 2014-2017 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -22,6 +22,8 @@
 #include "grn_normalizer.h"
 
 #include <string.h>
+#include <math.h>
+#include <float.h>
 
 #ifdef GRN_WITH_ONIGMO
 # define GRN_SUPPORT_REGEXP
@@ -125,6 +127,46 @@ grn_operator_to_string(grn_operator op)
   }
 }
 
+grn_operator_exec_func *
+grn_operator_to_exec_func(grn_operator op)
+{
+  grn_operator_exec_func *func = NULL;
+
+  switch (op) {
+  case GRN_OP_EQUAL :
+    func = grn_operator_exec_equal;
+    break;
+  case GRN_OP_NOT_EQUAL :
+    func = grn_operator_exec_not_equal;
+    break;
+  case GRN_OP_LESS :
+    func = grn_operator_exec_less;
+    break;
+  case GRN_OP_GREATER :
+    func = grn_operator_exec_greater;
+    break;
+  case GRN_OP_LESS_EQUAL :
+    func = grn_operator_exec_less_equal;
+    break;
+  case GRN_OP_GREATER_EQUAL :
+    func = grn_operator_exec_greater_equal;
+    break;
+  case GRN_OP_MATCH :
+    func = grn_operator_exec_match;
+    break;
+  case GRN_OP_PREFIX :
+    func = grn_operator_exec_prefix;
+    break;
+  case GRN_OP_REGEXP :
+    func = grn_operator_exec_regexp;
+    break;
+  default :
+    break;
+  }
+
+  return func;
+}
+
 #define DO_EQ_SUB do {\
   switch (y->header.domain) {\
   case GRN_DB_INT8 :\
@@ -155,7 +197,7 @@ grn_operator_to_string(grn_operator op)
     r = (x_ == GRN_UINT64_VALUE(y));\
     break;\
   case GRN_DB_FLOAT :\
-    r = ((x_ <= GRN_FLOAT_VALUE(y)) && (x_ >= GRN_FLOAT_VALUE(y)));\
+    r = (fabs(x_ - GRN_FLOAT_VALUE(y)) < DBL_EPSILON);\
     break;\
   case GRN_DB_SHORT_TEXT :\
   case GRN_DB_TEXT :\
